from .clients.bluesky_event_stream import BlueskyEventStream
from .clients.bluesky_run import BlueskyRun
from .clients.catalog_of_bluesky_runs import CatalogOfBlueskyRuns
from .writing.tiled_writer import TiledWriter

__all__ = [
    "BlueskyEventStream",
    "BlueskyRun",
    "CatalogOfBlueskyRuns",
    "TiledWriter",
]



import json

import uuid
from pathlib import Path
import jinja2


def render_templated_documents(fname: str, data_dir: str):
    dirpath = str(Path(__file__).parent)
    env = jinja2.Environment(loader=jinja2.FileSystemLoader(dirpath), autoescape=False)
    template = env.get_template(fname)
    rendered = template.render(root_path=data_dir, uuid=str(uuid.uuid4())[:-12])

    yield from json.loads(rendered)



import numpy as np
import pytest
from examples.render import render_templated_documents
from tiled.client.array import ArrayClient
from tiled.client.dataframe import DataFrameClient

from bluesky_tiled_plugins import TiledWriter
from bluesky_tiled_plugins.writing.validator import (
    validate_reading,
    validate_structure,
    validate,
    ReadingValidationException,
)


def test_validate_structure_shape(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with an introduced shape error
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the document to introduce an error
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["shape"] = [1, 2, 3]  # should be [1, 13, 17]

        tw(name, doc)  # Write the document

    # Get the client for the written data
    array_client = client[uid]["primary/det-key2"]
    assert isinstance(array_client, ArrayClient)

    # Try validating the structure
    with pytest.raises(ValueError, match="Shape mismatch"):
        validate_structure(array_client, fix_errors=False)

    # Now validate and fix the error
    notes = validate_structure(array_client, fix_errors=True)
    assert array_client.shape == (3, 13, 17)
    assert array_client.read().shape == (3, 13, 17)
    assert any("Fixed shape mismatch" in note for note in notes)


def test_validate_structure_chunks(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with an introduced chunk shape error
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the document to introduce an error
        if name in {"resource", "stream_resource"}:
            doc["parameters"]["chunk_shape"] = [1, 2, 3]  # should be [3, 13, 17]

        tw(name, doc)  # Write the document

    # Get the client for the written data
    array_client = client[uid]["primary/det-key2"]
    assert isinstance(array_client, ArrayClient)

    # Try validating the structure
    with pytest.raises(ValueError, match="Chunk shape mismatch"):
        validate_structure(array_client, fix_errors=False)

    # Now validate and fix the error
    notes = validate_structure(array_client, fix_errors=True)
    assert array_client.chunks == ((3,), (13,), (17,))
    assert array_client.read() is not None
    assert any("Fixed chunk shape mismatch" in note for note in notes)


def test_validate_structure_dtype(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with an introduced dtype error
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the document to introduce an error
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["dtype_numpy"] = np.dtype(
                "int32"
            ).str  # should be "int64"

        tw(name, doc)  # Write the document

    # Get the client for the written data
    array_client = client[uid]["primary/det-key2"]
    assert isinstance(array_client, ArrayClient)

    # Try validating the structure
    with pytest.raises(ValueError, match="Data type mismatch"):
        validate_structure(array_client, fix_errors=False)

    # Now validate and fix the error
    notes = validate_structure(array_client, fix_errors=True)
    assert array_client.dtype == np.dtype("int64")
    assert array_client.read() is not None
    assert any("Fixed dtype mismatch" in note for note in notes)


def test_validate_reading_array_success(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with TiledWriter
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        tw(name, doc)  # Write the document

    # Get the client for the written data
    array_client = client[uid]["primary/det-key2"]
    assert isinstance(array_client, ArrayClient)
    assert validate_reading(array_client) is None


def test_validate_reading_array_failure(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with TiledWriter, point to non-existent file
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Point to a non-existent file to simulate read error
        if name == "stream_resource":
            doc["uri"] += "_non_existent"

        tw(name, doc)  # Write the document

    # Get the client for the written data
    array_client = client[uid]["primary/det-key2"]
    assert isinstance(array_client, ArrayClient)

    # Validate reading should raise an exception
    with pytest.raises(ReadingValidationException):
        validate_reading(array_client)

    # Now validate reading with ignored errors
    assert (
        validate_reading(
            array_client, ignore_errors=["Unable to synchronously open file"]
        )
        is None
    )


def test_validate_reading_table_success(client):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with TiledWriter
    documents = render_templated_documents("internal_events.json", "")
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        tw(name, doc)  # Write the document

    # Get the client for the written data
    table_client = client[uid]["primary"].base["internal"]
    assert isinstance(table_client, DataFrameClient)
    assert validate_reading(table_client) is None


def test_validate_bluesky_run_success(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with TiledWriter
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        tw(name, doc)  # Write the document

    # Run the full validation
    assert validate(client[uid]) is True

    # There should be no validation notes since everything is correct
    assert client[uid].metadata.get("notes") is None


def test_validate_bluesky_run_failure(client, external_assets_folder):
    tw = TiledWriter(client, validate=False)  # Do not validate on write (default)

    # Write documents with an introduced shape error
    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the document to introduce an error
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["shape"] = [1, 2, 3]  # should be [1, 13, 17]

        tw(name, doc)  # Write the document

    # Run the full validation, which should fail/raise
    with pytest.raises(ValueError, match="Shape mismatch"):
        validate(client[uid], fix_errors=False, raise_on_error=True)

    assert validate(client[uid], fix_errors=False, raise_on_error=False) is False

    # Now run validation with fixing enabled
    assert validate(client[uid], fix_errors=True) is True

    # There should be validation notes about the fixes applied
    notes = client[uid].metadata.get("notes", [])
    assert any("Fixed shape mismatch" in note for note in notes)



from __future__ import annotations

import importlib.metadata
from pathlib import Path
from subprocess import check_output
from typing import Any

project = "bluesky-tiled-plugins"
copyright = "Bluesky Collaboration"
author = "Bluesky Collaboration"
github_user = "bluesky"

# The full version, including alpha/beta/rc tags.
release = importlib.metadata.version("bluesky_tiled_plugins")

# The short X.Y version.
if "+" in release:
    # Not on a tag, use branch name
    root = Path(__file__).absolute().parent.parent
    git_branch = check_output("git branch --show-current".split(), cwd=root)
    version = release = git_branch.decode().strip()
else:
    version = release

extensions = [
    "autodoc2",
    "myst_parser",
    "numpydoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinxcontrib.mermaid",
    "sphinx_copybutton",
    "sphinx_design",
]

source_suffix = [".rst", ".md"]
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    ".venv",
]

html_theme = "pydata_sphinx_theme"
html_logo = "_static/logo_bluesky.svg"
html_theme_options: dict[str, Any] = {
    "github_url": f"https://github.com/{github_user}/{project}",
    "external_links": [
        {
            "name": "Bluesky Project",
            "url": "https://blueskyproject.io",
        },
    ],
    "icon_links": [
        {
            "name": "PyPI",
            "url": f"https://pypi.org/project/{project}",
            "icon": "fas fa-cube",
        },
    ],
}
html_context = {
    "github_user": github_user,
    "github_repo": project,
    "github_version": version,
    "doc_path": "docs",
}

myst_enable_extensions = [
    "colon_fence",
]

intersphinx_mapping = {
    "bluesky": ("https://blueskyproject.io/bluesky/main", None),
    "event_model": ("https://blueskyproject.io/event-model/main", None),
    "numpy": ("https://numpy.org/devdocs/", None),
    "python": ("https://docs.python.org/3", None),
    "tiled": ("https://blueskyproject.io/tiled", None),
}

nitpick_ignore = [
    ("py:class", "_io.StringIO"),
    ("py:class", "_io.BytesIO"),
]
# Ignore some missing intersphinx (workaround)
nitpick_ignore_regex = [
    (r"py:.*", r"event_model\..*"),
    (r"py:.*", r"tiled\..*"),
    (r"py:.*", r"bluesky_tiled_plugins._version\..*"),
    (r"py:.*", r"bluesky_tiled_plugins.clients.bluesky_run._BlueskyRunSQL"),
]

always_document_param_types = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Set copy-button to ignore python and bash prompts
# https://sphinx-copybutton.readthedocs.io/en/latest/use.html#using-regexp-prompt-identifiers
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# Which package to load and document
autodoc2_packages = [{"path": "../src/bluesky_tiled_plugins"}]

# Put them in docs/_api which is git ignored
autodoc2_output_dir = "_api"

autodoc2_render_plugin = "myst"

# Don't document private things
autodoc2_hidden_objects = {"private", "dunder", "inherited"}

# We don't have any docstring for __init__, so by separating
# them here we don't get the "Initilize" text that would otherwise be added
autodoc2_class_docstring = "both"

# Which objects to include docstrings for. ‘direct’ means only from objects
# that are not inherited.
autodoc2_docstrings = "all"



import asyncio

import pytest
from bluesky.run_engine import RunEngine, TransitionError

import h5py
import copy
import tifffile as tf
import numpy as np
from bluesky_tiled_plugins.exporters import json_seq_exporter

rng = np.random.default_rng(12345)


@pytest.fixture(scope="function", params=[False, True])
def RE(request):
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    RE = RunEngine({}, call_returns_result=request.param, loop=loop)

    def clean_event_loop():
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass
        loop.call_soon_threadsafe(loop.stop)
        RE._th.join()
        loop.close()

    request.addfinalizer(clean_event_loop)
    return RE


@pytest.fixture(scope="module")
def catalog(tmp_path_factory):
    tiled_catalog = pytest.importorskip("tiled.catalog")
    tmp_path = tmp_path_factory.mktemp("tiled_catalog")
    return tiled_catalog.in_memory(
        writable_storage={
            "filesystem": str(tmp_path),
            "sql": f"duckdb:///{tmp_path}/test.db",
        },
        readable_storage=[str(tmp_path.parent)],
    )


@pytest.fixture(scope="module")
def app(catalog):
    tsa = pytest.importorskip("tiled.server.app")
    default_serialization_registry = pytest.importorskip(
        "tiled.media_type_registration"
    ).default_serialization_registry

    serialization_registry = copy.deepcopy(default_serialization_registry)
    serialization_registry.register(
        "BlueskyRun", "application/json-seq", json_seq_exporter
    )

    return tsa.build_app(catalog, serialization_registry=serialization_registry)


@pytest.fixture(scope="module")
def context(app):
    tc = pytest.importorskip("tiled.client")
    with tc.Context.from_app(app) as context:
        yield context


@pytest.fixture(scope="module")
def client(context):
    tc = pytest.importorskip("tiled.client")
    return tc.from_context(context)


@pytest.fixture(scope="module")
def external_assets_folder(tmp_path_factory):
    """External data files used with the saved documents."""
    # Create a temporary directory
    temp_dir = tmp_path_factory.mktemp("example_files")

    # Create an external hdf5 file
    with h5py.File(temp_dir.joinpath("dataset.h5"), "w") as file:
        grp = file.create_group("entry").create_group("data")
        grp.create_dataset("data_1", data=rng.random(size=(3,), dtype="float64"))
        grp.create_dataset(
            "data_2", data=rng.integers(-10, 10, size=(3, 13, 17)), dtype="<i8"
        )

    # Create a second external hdf5 file to be declared in a different stream resource
    with h5py.File(temp_dir.joinpath("dataset_part2.h5"), "w") as file:
        grp = file.create_group("entry").create_group("data")
        grp.create_dataset(
            "data_2", data=rng.integers(-10, 10, size=(5, 13, 17)), dtype="<i8"
        )

    # Create a sequence of tiff files
    (temp_dir / "tiff_files").mkdir(parents=True, exist_ok=True)
    for i in range(3):
        data = rng.integers(0, 255, size=(1, 10, 15), dtype="uint8")
        tf.imwrite(temp_dir.joinpath("tiff_files", f"img_{i:05}.tif"), data)

    return str(temp_dir.absolute()).replace("\\", "/")



import json
import os

import pytest

from bluesky_tiled_plugins.writing._json_writer import JSONLinesWriter, JSONWriter


def read_json_file(path):
    with open(path) as f:
        return json.load(f)


def read_jsonl_file(path):
    with open(path) as f:
        return [json.loads(line) for line in f if line.strip()]


@pytest.mark.parametrize(
    "writer_class, extension", [(JSONWriter, "json"), (JSONLinesWriter, "jsonl")]
)
def test_json_writer(tmpdir, writer_class, extension):
    writer = writer_class(tmpdir)
    start_doc = {"uid": "abc", "value": 1}
    event_doc = {"seq_num": 1, "data": {"x": 1}}
    stop_doc = {"exit_status": "success"}

    writer("start", start_doc)
    writer("event", event_doc)
    writer("stop", stop_doc)

    # Read the file and check its contents
    filename = os.path.join(tmpdir, f"abc.{extension}")
    data = (
        read_json_file(filename)
        if (writer_class == JSONWriter)
        else read_jsonl_file(filename)
    )
    assert data[0]["name"] == "start"
    assert data[1]["name"] == "event"
    assert data[2]["name"] == "stop"


@pytest.mark.parametrize(
    "writer_class, extension", [(JSONWriter, "json"), (JSONLinesWriter, "jsonl")]
)
def test_custom_filename(tmpdir, writer_class, extension):
    writer = writer_class(tmpdir, filename=f"custom.{extension}")
    doc = {"uid": "value"}
    writer("start", doc)
    assert os.path.exists(os.path.join(tmpdir, f"custom.{extension}"))



import orjson

from bluesky_tiled_plugins.utils import truncate_json_overflow


def test_truncate_json_overflow():
    # Test with a large integer
    data = {"large_pos_int": 2**60, "large_neg_int": -(2**60)}
    truncated_data = truncate_json_overflow(data)
    assert orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER)
    for val in orjson.loads(
        orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER)
    ).values():
        assert val is not None

    # Test with a large float
    data = {"large_pos_float": 2e308, "large_neg_float": -2e308}
    truncated_data = truncate_json_overflow(data)
    assert orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER)
    for val in orjson.loads(
        orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER)
    ).values():
        assert val is not None

    # Test with a list of large integers and floats
    data = [[2**60, -(2**60)], [2e308, -2e308]]
    truncated_data = truncate_json_overflow(data)
    assert orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER)

    # Test with a dictionary containing various types
    data = {
        "int": 42,
        "float": 3.14,
        "str": "Hello, world!",
        "list": [1, 2, 3],
        "dict": {"key": "value"},
        "large_int": 2**60,
        "large_float": 2e308,
        "nested": {
            "large_neg_int": -(2**60),
            "large_neg_float": -2e308,
            "list_of_large_ints": [2**60, -(2**60)],
            "list_of_large_floats": [2e308, -2e308],
        },
    }
    truncated_data = truncate_json_overflow(data)
    assert orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER)

    # Test with a NaN value
    data = {"nan": float("nan")}
    truncated_data = truncate_json_overflow(data)
    assert (
        orjson.loads(orjson.dumps(truncated_data, option=orjson.OPT_STRICT_INTEGER))[
            "nan"
        ]
        is None
    )



import json
import os
from collections.abc import Iterator
from typing import cast
from urllib.parse import parse_qs, urlparse

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import h5py
import numpy as np
import ophyd.sim
import pytest
import tifffile as tf
from bluesky.protocols import (
    Collectable,
    HasName,
    Readable,
    Reading,
    StreamAsset,
    WritesStreamAssets,
)
from event_model.documents.event_descriptor import DataKey
from event_model.documents.stream_datum import StreamDatum
from event_model.documents.stream_resource import StreamResource
from tiled.client import record_history
from examples.render import render_templated_documents

from bluesky_tiled_plugins import TiledWriter


class Named(HasName):
    name: str = ""
    root: str = "/root"
    parent = None

    def __init__(self, name: str, root: str) -> None:
        self.name = name
        self.root = root
        self.counter = 0


class StreamDatumReadableCollectable(Named, Readable, Collectable, WritesStreamAssets):
    """Produces no events, but only StreamResources/StreamDatums and can be read or collected"""

    def _get_hdf5_stream(
        self, data_key: str, index: int
    ) -> tuple[StreamResource | None, StreamDatum]:
        file_path = os.path.join(self.root, "dataset.h5")
        uid = f"{data_key}-uid"
        data_desc = self.describe()[
            data_key
        ]  # Descriptor dictionary for the current data key
        data_shape = cast("tuple[int, ...]", tuple(data_desc["shape"]))
        hdf5_dataset = f"/{data_key}/VALUE"

        stream_resource = None
        if self.counter == 0:
            # Backward compatibility test, ignore typing errors
            stream_resource = StreamResource(  # type: ignore[typeddict-unknown-key]
                parameters={
                    "dataset": hdf5_dataset,
                    "chunk_shape": (100, *data_shape[1:]),
                },
                data_key=data_key,
                root=self.root,
                resource_path="/dataset.h5",
                uri="file://localhost/" + file_path,
                spec="AD_HDF5_SWMR_STREAM",
                mimetype="application/x-hdf5",
                uid=uid,
            )
            # Initialize an empty HDF5 dataset (3D: var 1 dim, fixed 2 and 3 dims)
            with h5py.File(file_path, "a") as f:
                dset = f.require_dataset(
                    hdf5_dataset,
                    data_shape,
                    maxshape=(None, *data_shape[1:]),
                    dtype=np.dtype("float64"),
                    chunks=(100, *data_shape[1:]),
                )

        indx_min, indx_max = self.counter, self.counter + index
        stream_datum = StreamDatum(
            stream_resource=uid,
            descriptor="",
            uid=f"{uid}/{self.counter}",
            indices={"start": indx_min, "stop": indx_max},
            seq_nums={
                "start": 0,
                "stop": 0,
            },  # seq_nums will be overwritten by RunBundler
        )

        # Write (append to) the hdf5 dataset
        with h5py.File(file_path, "a") as f:
            dset = f[hdf5_dataset]
            dset.resize([indx_max * data_shape[0], *data_shape[1:]])
            dset[indx_min * data_shape[0] : indx_max * data_shape[0], ...] = (
                np.random.randn((indx_max - indx_min) * data_shape[0], *data_shape[1:])
            )

        return stream_resource, stream_datum

    def _get_tiff_stream(
        self, data_key: str, index: int
    ) -> tuple[StreamResource | None, StreamDatum]:
        file_path = self.root
        for data_key in [f"{self.name}-sd3"]:
            uid = f"{data_key}-uid"
            data_desc = self.describe()[
                data_key
            ]  # Descriptor dictionary for the current data key
            data_shape = cast("tuple[int, ...]", tuple(data_desc["shape"]))
            stream_resource = None
            if self.counter == 0:
                # Backward compatibility test, ignore typing errors
                stream_resource = StreamResource(  # type: ignore[typeddict-unknown-key]
                    parameters={
                        "chunk_shape": (1, *data_shape),
                        "template": "{:05d}.tif",
                        "join_method": "stack",
                    },
                    data_key=data_key,
                    root=self.root,
                    uri="file://localhost/" + self.root + "/",
                    spec="AD_TIFF",
                    mimetype="multipart/related;type=image/tiff",
                    uid=uid,
                )

            indx_min, indx_max = self.counter, self.counter + index
            stream_datum = StreamDatum(
                stream_resource=uid,
                descriptor="",
                uid=f"{uid}/{self.counter}",
                indices={"start": indx_min, "stop": indx_max},
                seq_nums={
                    "start": 0,
                    "stop": 0,
                },  # seq_nums will be overwritten by RunBundler
            )

            # Write a tiff file
            data = np.random.randint(0, 255, data_shape, dtype="uint8")
            tf.imwrite(os.path.join(file_path, f"{self.counter:05}.tif"), data)

        return stream_resource, stream_datum

    def describe(self) -> dict[str, DataKey]:
        """Describe datasets which will be backed by StreamResources"""
        return {
            # Numerical data with 1 number per event in hdf5 format
            f"{self.name}-sd1": DataKey(
                source="file",
                dtype="number",
                dtype_numpy=np.dtype("float64").str,
                shape=[
                    1,
                ],
                external="STREAM:",
            ),
            # 2-D data with 5 frames per event in hdf5 format
            f"{self.name}-sd2": DataKey(
                source="file",
                dtype="array",
                dtype_numpy=np.dtype("float64").str,
                shape=[5, 10, 15],
                external="STREAM:",
            ),
            # 3-D data with 10 frames per event in tiff format
            f"{self.name}-sd3": DataKey(
                source="file",
                dtype="array",
                dtype_numpy=np.dtype("uint8").str,
                shape=[10, 5, 7, 4],
                external="STREAM:",
            ),
        }

    def describe_collect(self) -> dict[str, DataKey] | dict[str, dict[str, DataKey]]:
        return self.describe()

    def collect_asset_docs(self, index: int | None = None) -> Iterator[StreamAsset]:
        """Produce a StreamResource and StreamDatum for all data keys for 0:index"""
        index = index or 1
        data_keys_methods = {
            f"{self.name}-sd1": self._get_hdf5_stream,
            f"{self.name}-sd2": self._get_hdf5_stream,
            f"{self.name}-sd3": self._get_tiff_stream,
        }

        for data_key, method in data_keys_methods.items():
            stream_resource, stream_datum = method(data_key, index)
            if stream_resource is not None:
                yield "stream_resource", stream_resource
            yield "stream_datum", stream_datum

        self.counter += index

    def get_index(self) -> int:
        """Report how many frames were written"""
        return self.counter

    def read(self) -> dict[str, Reading]:
        """Produce an empty event"""
        return {}


class SynSignalWithRegistry(ophyd.sim.SynSignalWithRegistry):
    """A readable image detector that writes a sequence of files and generates relevant Bluesky documents.

    Subclassed from ophyd.sim to match the updated schema of Resource documents.
    """

    def __init__(self, *args, dtype_numpy="uint8", **kwargs):
        self.dtype_numpy = dtype_numpy
        super().__init__(*args, **kwargs)

    def stage(self):
        super().stage()
        parameters = {
            "chunk_shape": (1,),
            "template": "_{:d}." + self.save_ext,
            "join_method": "stack",
        }
        self._asset_docs_cache[-1][1]["resource_kwargs"].update(parameters)

    def describe(self):
        res = super().describe()
        for key in res:
            res[key]["external"] = "FILESTORE"
            res[key]["dtype_numpy"] = self.dtype_numpy
        return res


def test_stream_datum_readable_counts(RE, client, tmp_path):
    tw = TiledWriter(client)
    det = StreamDatumReadableCollectable(name="det", root=str(tmp_path))
    RE(bp.count([det], 3), tw)
    stream = client.values().last()["primary"]
    keys = sorted(set(stream.base.keys()).difference({"internal"}))

    assert stream[keys[0]].shape == (3,)
    assert stream[keys[1]].shape == (15, 10, 15)
    assert stream[keys[2]].shape == (3, 10, 5, 7, 4)
    assert stream[keys[0]].read() is not None
    assert stream[keys[1]].read() is not None
    assert stream[keys[2]].read() is not None


def test_stream_datum_readable_with_two_detectors(RE, client, tmp_path):
    det1 = StreamDatumReadableCollectable(name="det1", root=str(tmp_path))
    det2 = StreamDatumReadableCollectable(name="det2", root=str(tmp_path))
    tw = TiledWriter(client)
    RE(bp.count([det1, det2], 3), tw)
    stream = client.values().last()["primary"]
    keys = sorted(set(stream.base.keys()).difference({"internal"}))

    assert stream[keys[0]].shape == (3,)
    assert stream[keys[1]].shape == (15, 10, 15)
    assert stream[keys[2]].shape == (3, 10, 5, 7, 4)
    assert stream[keys[3]].shape == (3,)
    assert stream[keys[4]].shape == (15, 10, 15)
    assert stream[keys[5]].shape == (3, 10, 5, 7, 4)
    assert stream[keys[0]].read() is not None
    assert stream[keys[1]].read() is not None
    assert stream[keys[2]].read() is not None
    assert stream[keys[3]].read() is not None
    assert stream[keys[4]].read() is not None
    assert stream[keys[5]].read() is not None


def test_stream_datum_collectable(RE, client, tmp_path):
    det = StreamDatumReadableCollectable(name="det", root=str(tmp_path))
    tw = TiledWriter(client)
    RE(collect_plan(det, name="primary"), tw)
    stream = client.values().last()["primary"]
    keys = sorted(set(stream.base.keys()).difference({"internal"}))

    assert stream[keys[0]].read() is not None
    assert stream[keys[1]].read() is not None
    assert stream[keys[2]].read() is not None


@pytest.mark.parametrize("frames_per_event", [1, 5, 10])
def test_handling_non_stream_resource(RE, client, tmp_path, frames_per_event):
    det = SynSignalWithRegistry(
        func=lambda: np.random.randint(
            0, 255, (frames_per_event, 10, 15), dtype="uint8"
        ),
        dtype_numpy=np.dtype("uint8").str,
        name="img",
        labels={"detectors"},
        save_func=tf.imwrite,
        save_path=str(tmp_path),
        save_spec="AD_TIFF",
        save_ext="tif",
    )
    tw = TiledWriter(client)
    RE(bp.count([det], 3), tw)
    extr = client.values().last()["primary"].base["img"]
    intr = client.values().last()["primary"].base["internal"]
    assert extr.shape == (3, frames_per_event, 10, 15)
    assert extr.read() is not None
    assert set(intr.columns) == {"seq_num", "time"}
    assert len(intr.read()) == 3
    assert (intr["seq_num"].read() == [1, 2, 3]).all()


def collect_plan(*objs, name="primary"):
    yield from bps.open_run()
    yield from bps.declare_stream(*objs, collect=True, name=name)
    yield from bps.collect(*objs, return_payload=False, name=name)
    yield from bps.close_run()


@pytest.mark.parametrize(
    "fname", ["internal_events", "external_assets", "external_assets_legacy"]
)
@pytest.mark.parametrize("batch_size", [0, 1, 1000, None])
def test_with_correct_sample_runs(client, batch_size, external_assets_folder, fname):
    if batch_size is None:
        tw = TiledWriter(client)
    else:
        tw = TiledWriter(client, batch_size=batch_size)
    for item in render_templated_documents(fname + ".json", external_assets_folder):
        if item["name"] == "start":
            uid = item["doc"]["uid"]
        tw(**item)

    run = client[uid]

    for stream in run.values():
        assert stream.read() is not None


def test_dims_names(client, external_assets_folder):
    tw = TiledWriter(client)

    for item in render_templated_documents(
        "external_assets.json", external_assets_folder
    ):
        if item["name"] == "start":
            uid = item["doc"]["uid"]
        tw(**item)

    run = client[uid]

    assert run["primary"]["det-key1"].structure().dims is None
    assert run["primary"]["det-key2"].structure().dims == ("time", "dim_x", "dim_y")


@pytest.mark.parametrize(
    "batch_size, expected_patch_shapes, expected_patch_offsets",
    [(1, (1, 1, 1), (0, 1, 2)), (2, (2, 1), (0, 2)), (5, (3,), (0,))],
)
def test_data_source_patching(
    client,
    batch_size,
    expected_patch_shapes,
    expected_patch_offsets,
    external_assets_folder,
):
    tw = TiledWriter(client, batch_size=batch_size)

    with record_history() as history:
        for item in render_templated_documents(
            "external_assets.json", external_assets_folder
        ):
            tw(**item)

    def parse_data_source_uri(uri: str):
        """Given a full data_source URL, extract:
            - data_key (e.g. "det-key1")
            - decoded query parameters as tuples of ints

        Returns:
            (data_key, params_dict)
        """

        # data_key is the last component of the path
        data_key = urlparse(uri).path.rstrip("/").split("/")[-1]

        # parse query parameters and convert comma-separated values to tuples of ints
        params = {}
        for k, v in parse_qs(urlparse(uri).query).items():
            params[k] = tuple(map(int, v[0].split(",")))  # parse_qs gives lists

        return data_key, params

    put_uri_params = [
        parse_data_source_uri(str(req.url))
        for req in history.requests
        if req.method == "PUT" and "/data_source" in req.url.path
    ]

    # Check that each data key received the expected number of updates
    assert len(put_uri_params) == 3 * len(
        expected_patch_shapes
    )  # 3 data keys in the example
    for data_key in ("det-key1", "det-key2", "det-key3"):
        assert len([uri for dk, uri in put_uri_params if dk == data_key]) == len(
            expected_patch_shapes
        )

        # Check that the patch sizes and offsets (leftmost dimensions) match expectations
        actual_patch_sizes = tuple(
            params["patch_shape"][0] for dk, params in put_uri_params if dk == data_key
        )
        actual_patch_offsets = tuple(
            params["patch_offset"][0] for dk, params in put_uri_params if dk == data_key
        )
        assert actual_patch_sizes == expected_patch_shapes
        assert actual_patch_offsets == expected_patch_offsets


@pytest.mark.parametrize("error_type", ["shape", "chunks", "dtype"])
@pytest.mark.parametrize("validate", [True, False])
def test_validate_external_data(client, external_assets_folder, error_type, validate):
    tw = TiledWriter(client, validate=validate)

    documents = render_templated_documents(
        "external_assets_single_key.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the document to introduce an error
        if (error_type == "shape") and (name == "descriptor"):
            doc["data_keys"]["det-key2"]["shape"] = [1, 2, 3]  # should be [1, 13, 17]
        elif (error_type == "chunks") and name in {"resource", "stream_resource"}:
            doc["parameters"]["chunk_shape"] = [1, 2, 3]  # should be [3, 13, 17]
        elif (error_type == "dtype") and (name == "descriptor"):
            doc["data_keys"]["det-key2"]["dtype_numpy"] = np.dtype(
                "int32"
            ).str  # should be "int64"

        # Check that the warning is issued when data changes during the validation
        if name == "stop" and validate:
            with pytest.warns(UserWarning):
                tw(name, doc)
        else:
            tw(name, doc)

    # Try reading the imported data
    run = client[uid]
    if not validate and not error_type == "chunks":
        with pytest.raises(ValueError):
            assert run["primary"].read() is not None
    else:
        assert run["primary"].read() is not None
        assert run["primary"]["det-key2"].read().shape == (3, 13, 17)


@pytest.mark.parametrize("squeeze", [True, False])
def test_slice_and_squeeze(client, external_assets_folder, squeeze):
    tw = TiledWriter(client)

    documents = render_templated_documents(
        "external_assets_single_key_two_files.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the documents to add slice and squeeze parameters
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["shape"] = [1, 17] if squeeze else [1, 5, 17]
        elif name in {"resource", "stream_resource"}:
            doc["parameters"]["slice"] = ":,5,:" if squeeze else ":,:5,:"
            doc["parameters"]["squeeze"] = squeeze
            doc["parameters"]["chunk_shape"] = [1]

        tw(name, doc)

    # Try reading the imported data
    assert client[uid]["primary"].read() is not None


def test_legacy_multiplier_parameter(client, external_assets_folder):
    tw = TiledWriter(client)

    documents = render_templated_documents(
        "external_assets_single_key_two_files.json", external_assets_folder
    )
    for item in documents:
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Modify the documents to add slice and squeeze parameters
        if name == "descriptor":
            doc["data_keys"]["det-key2"]["shape"] = [13, 17]
        elif name in {"resource", "stream_resource"}:
            doc["parameters"]["multiplier"] = 1

        tw(name, doc)

    # Try reading the imported data
    assert client[uid]["primary"].read() is not None


def test_streams_with_no_events(client, external_assets_folder):
    tw = TiledWriter(client)

    for item in render_templated_documents(
        "external_assets_single_key_two_files.json", external_assets_folder
    ):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        # Skip the resource and datum documents
        if name in {"resource", "stream_resource", "datum", "stream_datum", "event"}:
            continue

        tw(name, doc)

    # Try reading the data -- should return an empty dataset
    assert client[uid]["primary"].read() is not None
    assert client[uid]["primary"].read().data_vars == {}
    assert client[uid]["primary"].metadata is not None


@pytest.mark.parametrize("include_data_sources", [True, False])
@pytest.mark.parametrize(
    "fname", ["internal_events", "external_assets", "external_assets_legacy"]
)
def test_zero_gets(client, external_assets_folder, fname, include_data_sources):
    client = client.new_variation(include_data_sources=include_data_sources)
    assert client._include_data_sources == include_data_sources
    tw = TiledWriter(client)
    assert bool(tw.client._include_data_sources)

    with record_history() as history:
        for item in render_templated_documents(fname + ".json", external_assets_folder):
            tw(**item)

    # Count the number of GET requests
    num_gets = sum(1 for req in history.requests if req.method == "GET")
    assert num_gets == 0


def test_bad_document_order(client, external_assets_folder):
    """Test that the TiledWriter can handle documents in a different order than expected

    Emit datum documents in the end, before the Stop document, but after corresponding Event documents.
    """
    tw = TiledWriter(client)

    document_cache = []
    for item in render_templated_documents(
        "external_assets_legacy.json", external_assets_folder
    ):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]

        if name == "datum":
            document_cache.append({"name": name, "doc": doc})
            continue

        if name == "stop":
            for cached_item in document_cache:
                tw(**cached_item)

        tw(**item)

    run = client[uid]

    for stream in run.values():
        assert stream.read() is not None
        assert "time" in stream.keys()
        assert "seq_num" in stream.keys()
        assert (
            len(stream.keys()) > 2
        )  # There's at least one data key in addition to time and seq_num


def test_json_backup(client, tmpdir, monkeypatch):
    def patched_event(name, doc):
        raise RuntimeError("This is a test error to check the backup functionality")

    monkeypatch.setattr(
        "bluesky_tiled_plugins.writing.tiled_writer._RunWriter.event", patched_event
    )

    tw = TiledWriter(client, backup_directory=str(tmpdir))

    for item in render_templated_documents("internal_events.json", ""):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]
        tw(**item)

    run = client[uid]

    assert (
        "primary" in run
    )  # The Descriptor was processed and the primary stream was created
    assert run["primary"].read() is not None  # The stream can be read
    assert len(run["primary"].read()) == 0  # No events were processed due to the error
    assert "stop" in run.metadata  # The TiledWriter did not crash

    # Check that the backup file was created
    filepath = tmpdir / f"{uid[:8]}.jsonl"
    assert filepath.exists()
    with open(filepath) as f:
        lines = [json.loads(line) for line in f if line.strip()]
    assert len(lines) == 7
    assert lines[0]["name"] == "start"
    assert lines[1]["name"] == "descriptor"
    assert lines[2]["name"].startswith("event")
    assert lines[6]["name"] == "stop"


@pytest.mark.parametrize(
    "max_array_size, expected_scheme",
    [(0, "file"), (4, "file"), (16, "duckdb"), (-1, "duckdb")],
)
def test_internal_arrays_written_as_zarr(client, max_array_size, expected_scheme):
    tw = TiledWriter(client, max_array_size=max_array_size)

    for item in render_templated_documents("internal_events.json", ""):
        name, doc = item["name"], item["doc"]
        if name == "start":
            uid = doc["uid"]
        tw(**item)

    run = client[uid]

    assert run["primary"]["long"].shape == (3, 8)
    assert run["primary"]["long"].read() is not None

    # There's a table and it is stored in the SQL database
    internal_table = run["primary"].base["internal"]
    assert internal_table.read() is not None
    assert (
        urlparse(internal_table.data_sources()[0].assets[0].data_uri).scheme == "duckdb"
    )

    if expected_scheme == "file":
        assert (
            "long" in run["primary"].base
        )  # There's a separate node for the array data
        assert (
            "long" not in internal_table.columns
        )  # The internal table does not have a column for it
        assert (
            urlparse(run["primary"]["long"].data_sources()[0].assets[0].data_uri).scheme
            == "file"
        )
    else:
        assert "long" not in run["primary"].base
        assert "long" in internal_table.columns
        assert run["primary"]["long"].data_sources() is None



import collections.abc


def truncate_json_overflow(data):
    """Truncate large numerical values to avoid overflow issues when serializing as JSON.

    This preemptively truncates large integers and floats with zero fractional part to fit within
    the JSON limits for integers, i.e. (-2^53, 2^53 - 1], in case the values are implicitly
    converted during serialization.
    """
    if isinstance(data, collections.abc.Mapping):
        return {k: truncate_json_overflow(v) for k, v in data.items()}
    if isinstance(data, collections.abc.Iterable) and not isinstance(data, str):
        # Handle lists, tuples, arrays, etc., but not strings
        return [truncate_json_overflow(item) for item in data]
    if (
        isinstance(data, (int, float))
        and not (data % 1)
        and not (1 - 2**53 <= data <= 2**53 - 1)
    ):
        return min(
            max(data, 1 - 2**53), 2**53 - 1
        )  # Truncate integers to fit in JSON (53 bits max)
    if isinstance(data, float) and (data < -1.7976e308 or data > 1.7976e308):
        return min(
            max(data, -1.7976e308), 1.7976e308
        )  # (Approx.) truncate floats to fit in JSON to avoid inf
    return data



import logging
import re
import time

from tiled.client.array import ArrayClient
from tiled.client.dataframe import DataFrameClient
from tiled.client.utils import handle_error, retry_context
from tiled.mimetypes import DEFAULT_ADAPTERS_BY_MIMETYPE as ADAPTERS_BY_MIMETYPE
from tiled.utils import safe_json_dump

logger = logging.getLogger(__name__)


class ValidationException(Exception):
    def __init__(self, message, uid=None):
        super().__init__(message)
        self.uid = uid


class ReadingValidationException(ValidationException):
    pass


class RunValidationException(ValidationException):
    pass


class MetadataValidationException(ValidationException):
    pass


def validate(
    root_client,
    fix_errors=True,
    try_reading=True,
    raise_on_error=False,
    ignore_errors=[],
):
    """Validate the given BlueskyRun client for completeness and data integrity.

    Parameters
    ----------

    root_client : tiled.client.run.RunClient
        The Run client to validate.
    fix_errors : bool, optional
        Whether to attempt to fix structural errors found during validation.
        Default is True.
    try_reading : bool, optional
        Whether to attempt reading the data for external data keys.
        Default is True.
    raise_on_error : bool, optional
        Whether to raise an exception on the first validation error encountered.
        Default is False.
    ignore_errors : list of str, optional
        List of error messages to ignore during reading validation.
        Default is an empty list.

    Returns
    -------
    bool
        True if validation passed without errors, False otherwise.
    """

    # Check if there's a Stop document in the run
    if "stop" not in root_client.metadata:
        logger.error("The Run is not complete: missing the Stop document")
        if raise_on_error:
            raise RunValidationException("Missing Stop document in the run")

    # Check all streams and data keys
    errored_keys, notes = [], []
    streams_node = (
        root_client["streams"] if "streams" in root_client.keys() else root_client
    )
    for sname, stream in streams_node.items():
        for data_key in stream.base:
            if data_key == "internal":
                continue

            data_client = stream[data_key]
            if data_client.data_sources()[0].management != "external":
                continue

            # Validate data structure
            title = f"Validation of data key '{sname}/{data_key}'"
            try:
                _notes = validate_structure(data_client, fix_errors=fix_errors)
                notes.extend([title + ": " + note for note in _notes])
            except Exception as e:
                msg = (
                    f"{type(e).__name__}: "
                    + str(e).replace("\n", " ").replace("\r", "").strip()
                )
                msg = title + f" failed with error: {msg}"
                logger.error(msg)
                if raise_on_error:
                    raise e
                notes.append(msg)

            # Validate reading of the data
            if try_reading:
                try:
                    validate_reading(data_client, ignore_errors=ignore_errors)
                except Exception as e:
                    errored_keys.append((sname, data_key, str(e)))
                    logger.error(
                        f"Reading validation of '{sname}/{data_key}' failed with error: {e}"
                    )
                    if raise_on_error:
                        raise e

            time.sleep(0.1)

    if try_reading and (not errored_keys):
        logger.info("Reading validation completed successfully.")

    # Update the root metadata with validation notes
    if notes:
        existing_notes = root_client.metadata.get("notes", [])
        root_client.update_metadata(
            {"notes": existing_notes + notes}, drop_revision=True
        )

    return not errored_keys


def validate_reading(data_client, ignore_errors=[]):
    """Attempt to read data from the given data client to validate data accessibility

    Parameters
    ----------
        data_client : tiled.client.ArrayClient or tiled.client.DataFrameClient
            The data client to validate reading from.
        ignore_errors : list of str, optional
            List of error messages to ignore during reading validation.
            Default is an empty list.

    Raises
    ------
        ReadingValidationException
            If reading the data fails with an unignored error.
    """

    data_key = data_client.item["id"]
    sname = data_client.item["attributes"]["ancestors"][-1]  # stream name

    if isinstance(data_client, ArrayClient):
        try:
            # Try to read the first and last elements
            idx_left_top = (0,) * len(data_client.shape)
            data_client[idx_left_top]
            idx_right_bottom = (-1,) * len(data_client.shape)
            data_client[idx_right_bottom]
        except Exception as e:
            if any([re.search(msg, str(e)) for msg in ignore_errors]):
                logger.info(f"Ignoring array reading error: {sname}/{data_key}: {e}")
            else:
                raise ReadingValidationException(
                    f"Array reading failed with error: {e}"
                )

    elif isinstance(data_client, DataFrameClient):
        try:
            data_client.read()  # try to read the entire table
        except Exception as e:
            if any([re.search(msg, str(e)) for msg in ignore_errors]):
                logger.info(f"Ignoring table reading error: {sname}/{data_key}: {e}")
            else:
                raise ReadingValidationException(
                    f"Table reading failed with error: {e}"
                )

    else:
        logger.warning(
            f"Validation of '{data_key=}' is not supported with client of type {type(data_client)}."
        )


def validate_structure(data_client, fix_errors=False) -> list[str]:
    """Validate and optionally fix the structure of the given (array) dataset.

    Parameters
    ----------
        data_client : tiled.client.ArrayClient
            The data client whose structure is to be validated.
        fix_errors : bool, optional
            Whether to attempt to fix structural errors found during validation.
            Default is False.

    Returns
    -------
        list of str
            A list of human-readable notes describing any fixes applied during validation.
    """

    data_source = data_client.data_sources()[0]
    uris = [asset.data_uri for asset in data_source.assets]
    structure = data_client.structure()
    notes = []

    # Initialize adapter from uris and determine the structure as read by the adapter
    adapter_class = ADAPTERS_BY_MIMETYPE[data_source.mimetype]
    true_structure = adapter_class.from_uris(
        *uris, **data_source.parameters
    ).structure()
    true_data_type = true_structure.data_type
    true_shape = true_structure.shape
    true_chunks = true_structure.chunks

    # Validate structure components
    if structure.shape != true_shape:
        if not fix_errors:
            raise ValueError(f"Shape mismatch: {structure.shape} != {true_shape}")
        else:
            msg = f"Fixed shape mismatch: {structure.shape} -> {true_shape}"
            logger.warning(msg)
            structure.shape = true_shape
            notes.append(msg)

    if structure.chunks != true_chunks:
        if not fix_errors:
            raise ValueError(
                f"Chunk shape mismatch: {structure.chunks} != {true_chunks}"
            )
        else:
            _true_chunk_shape = tuple(c[0] for c in true_chunks)
            _chunk_shape = tuple(c[0] for c in structure.chunks)
            msg = f"Fixed chunk shape mismatch: {_chunk_shape} -> {_true_chunk_shape}"
            logger.warning(msg)
            structure.chunks = true_chunks
            notes.append(msg)

    if structure.data_type != true_data_type:
        if not fix_errors:
            raise ValueError(
                f"Data type mismatch: {structure.data_type} != {true_data_type}"
            )
        else:
            msg = f"Fixed dtype mismatch: {structure.data_type.to_numpy_dtype()} -> {true_data_type.to_numpy_dtype()}"  # noqa
            logger.warning(msg)
            structure.data_type = true_data_type
            notes.append(msg)

    if structure.dims and (len(structure.dims) != len(true_shape)):
        if not fix_errors:
            raise ValueError(
                f"Number of dimension names mismatch for a {len(true_shape)}-dimensional array: {structure.dims}"
            )  # noqa
        else:
            old_dims = structure.dims
            if len(old_dims) < len(true_shape):
                structure.dims = (
                    ("time",)
                    + old_dims
                    + tuple(
                        f"dim{i}" for i in range(len(old_dims) + 1, len(true_shape))
                    )
                )
            else:
                structure.dims = old_dims[: len(true_shape)]
            msg = f"Fixed dimension names: {old_dims} -> {structure.dims}"
            logger.warning(msg)
            notes.append(msg)

    # Update the data source structure if any fixes were applied
    if notes:
        data_source.structure = structure
        for attempt in retry_context():
            with attempt:
                response = data_client.context.http_client.put(
                    data_client.uri.replace(
                        "/api/v1/metadata/", "/api/v1/data_source/", 1
                    ),
                    content=safe_json_dump({"data_source": data_source}),
                )
        handle_error(response)

    return notes



import pytest
from examples.render import render_templated_documents

from bluesky_tiled_plugins import TiledWriter


@pytest.fixture(scope="module", params=["internal_events", "external_assets"])
def run_client(client, external_assets_folder, request):
    tw = TiledWriter(client)
    for item in render_templated_documents(
        request.param + ".json", external_assets_folder
    ):
        if item["name"] == "start":
            uid = item["doc"]["uid"]
        tw(**item)

    yield client[uid]


def test_documents(run_client):
    assert len(list(run_client.v3.documents())) > 0
    assert len(list(run_client.v2.documents())) > 0



import copy
import functools
import io
import json
import keyword
import warnings
from datetime import datetime

from tiled.client.container import Container
from tiled.client.utils import handle_error, retry_context

from ._common import IPYTHON_METHODS
from .bluesky_event_stream import BlueskyEventStreamV2SQL
from .document import (
    DatumPage,
    Descriptor,
    Event,
    EventPage,
    Resource,
    Start,
    Stop,
    StreamDatum,
    StreamResource,
)

_document_types = {
    "start": Start,
    "stop": Stop,
    "event": Event,
    "descriptor": Descriptor,
    "event_page": EventPage,
    "datum_page": DatumPage,
    "resource": Resource,
    "stream_resource": StreamDatum,
    "stream_datum": StreamResource,
}


class BlueskyRun(Container):
    _ipython_display_ = None
    _repr_mimebundle_ = None

    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting from BlueskyRun, return the class itself
        if cls is not BlueskyRun:
            return super().__new__(cls)

        # Set the version based on the specs
        _cls = BlueskyRunV3 if cls._is_sql(item) else BlueskyRunV2Mongo
        return _cls(context, item=item, structure_clients=structure_clients, **kwargs)

    @staticmethod
    def _is_sql(item):
        for spec in item["attributes"]["specs"]:
            if spec["name"] == "BlueskyRun":
                if spec["version"].startswith("3."):
                    return True
                return False

    def __repr__(self):
        metadata = self.metadata
        datetime_ = datetime.fromtimestamp(metadata["start"]["time"])
        return (
            f"<BlueskyRun v{self._version} "
            f"{set(self)!r} "  # show the keys
            f"scan_id={metadata['start'].get('scan_id', 'UNSET')!s} "  # (scan_id is optional in the schema)
            f"uid={metadata['start']['uid'][:8]!r} "  # truncated uid
            f"{datetime_.isoformat(sep=' ', timespec='minutes')}"
            ">"
        )

    @property
    def start(self):
        """
        The Run Start document. A convenience alias:

        >>> run.start is run.metadata["start"]
        True
        """
        return self.metadata["start"]

    @property
    def stop(self):
        """
        The Run Stop document. A convenience alias:

        >>> run.stop is run.metadata["stop"]
        True
        """
        return self.metadata["stop"]

    @functools.cached_property
    def descriptors(self):
        return [doc for name, doc in self.documents() if name == "descriptor"]

    def __getattr__(self, key):
        """
        Let run.X be a synonym for run['X'] unless run.X already exists.

        This behavior is the same as with pandas.DataFrame.
        """
        # The wisdom of this kind of "magic" is arguable, but we
        # need to support it for backward-compatibility reasons.
        if key in IPYTHON_METHODS:
            raise AttributeError(key)
        if key in self:
            return self[key]
        raise AttributeError(key)

    def __dir__(self):
        # Build a list of entries that are valid attribute names
        # and add them to __dir__ so that they tab-complete.
        tab_completable_entries = [
            entry
            for entry in self
            if (entry.isidentifier() and (not keyword.iskeyword(entry)))
        ]
        return super().__dir__() + tab_completable_entries

    def describe(self):
        "For back-compat with intake-based BlueskyRun"
        warnings.warn(
            "This will be removed. Use .metadata directly instead of describe()['metadata'].",
            DeprecationWarning,
            stacklevel=2,
        )
        return {"metadata": self.metadata}

    def __call__(self):
        warnings.warn(
            "Do not call a BlueskyRun. For now this returns self, for "
            "backward-compatibility. but it will be removed in a future "
            "release.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self

    def read(self):
        raise NotImplementedError(
            "Reading any entire run is not supported. Access a stream in this run and read that."
        )

    @property
    def base(self):
        "Return the base Container client instead of a BlueskyRun client"
        return Container(
            self.context,
            item=self.item,
            structure_clients=self.structure_clients,
            queries=self._queries,
            sorting=self._sorting,
            include_data_sources=self._include_data_sources,
        )

    to_dask = read


class BlueskyRunV2(BlueskyRun):
    """A MongoDB-native layout of BlueskyRuns

    This layout has been in use prior to the introduction of SQL backend in May 2025.
    """

    _version = "2.0"

    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting, return the class itself
        if cls is not BlueskyRunV2:
            return super().__new__(
                cls, context, item=item, structure_clients=structure_clients, **kwargs
            )

        _cls = BlueskyRunV2SQL if cls._is_sql(item) else BlueskyRunV2Mongo
        return _cls(context, item=item, structure_clients=structure_clients, **kwargs)

    @property
    def v1(self):
        "Accessor to legacy interface."
        from databroker.v1 import Broker, Header

        db = Broker(self)
        header = Header(self, db)
        return header

    @property
    def v2(self):
        return self

    @property
    def v3(self):
        if not self._is_sql(self.item):
            raise NotImplementedError(
                "v3 is not available for MongoDB-based BlueskyRun"
            )

        structure_clients = copy.copy(self.structure_clients)
        structure_clients.set("BlueskyRun", lambda: BlueskyRunV3)
        return BlueskyRunV3(
            self.context, item=self.item, structure_clients=structure_clients
        )


class BlueskyRunV2Mongo(BlueskyRunV2):
    def documents(self, fill=False):
        if fill == "yes":
            fill = True
        elif fill == "no":
            fill = False
        elif fill == "delayed":
            raise NotImplementedError("fill='delayed' is not supported")
        else:
            fill = bool(fill)
        link = self.item["links"]["self"].replace("/metadata", "/documents", 1)
        for attempt in retry_context():
            with attempt:
                with self.context.http_client.stream(
                    "GET",
                    link,
                    params={"fill": fill},
                    headers={"Accept": "application/json-seq"},
                ) as response:
                    if response.is_error:
                        response.read()
                        handle_error(response)
                    tail = ""
                    for chunk in response.iter_bytes():
                        for line in chunk.decode().splitlines(keepends=True):
                            if line[-1] == "\n":
                                item = json.loads(tail + line)
                                yield (
                                    item["name"],
                                    _document_types[item["name"]](item["doc"]),
                                )
                                tail = ""
                            else:
                                tail += line
                    if tail:
                        item = json.loads(tail)
                        yield (item["name"], _document_types[item["name"]](item["doc"]))


class _BlueskyRunSQL(BlueskyRun):
    """A base class for a BlueskyRun that is backed by a SQL database.

    This class implements the SQL-specific method for accessing the stream of
    Bluesky documents. It is not intended to be used directly, but rather as a
    base class for other classes (v2 and v3) that implement additional methods.
    """

    @functools.cached_property
    def _has_streams_namespace(self) -> bool:
        """Determine whether the BlueskyRun has an intermediate "streams" namespace.

        Maintained for backward compatibility. Returns True if the following conditions are met:
        1. There is a "streams" key in the base container.
        2. The specs of the "streams" container do not include "BlueskyEventStream",
           indicating that "streams" is not itself a BlueskyEventStream.
        """
        return ("streams" in self.base) and (
            "BlueskyEventStream" not in {s.name for s in self.base["streams"].specs}
        )

    @functools.cached_property
    def _stream_names(self) -> list[str]:
        """Get the sorted list of stream names in the BlueskyRun.

        This property accounts for both the new layout (without "streams" namespace)
        and the old layout (with "streams" namespace), in which case the stream names
        are derived from the keys under the "streams" namespace.
        """

        return sorted(
            k
            for k in (
                self.base["streams"] if self._has_streams_namespace else self.base
            )
        )

    def __getitem__(self, key):
        if isinstance(key, tuple):
            key = "/".join(key)

        base_class = super()  # The base Container class

        def _base_getitem(key):
            # Try to get the item directly from the new container layout. Consider nested keys.
            try:
                return base_class.__getitem__(key)
            except KeyError as e:
                try:
                    # The requested key might be a column in the "internal" table
                    key = key.split("/")
                    key.insert(-1, "internal")
                    return base_class.__getitem__("/".join(key))
                except KeyError:
                    raise KeyError(
                        f"Key '{key[-1]}' not found in the BlueskyRun container"
                    ) from e

        # Back-compatibility for old versions of BlueskyRun layout that included 'streams' namespace.
        # This takes into account the possibility of an actual BlueskyEventStream to be named 'streams'.
        try:
            return _base_getitem(key)
        except KeyError as e:
            if key == "streams":
                warnings.warn(
                    "Looks like you are trying to access the 'streams' namespace, "
                    "but there is no 'streams' namespace in this BlueskyRun, which follows the new layout. "
                    "Please use the stream names directly, e.g. run['primary'] instead of run['streams/primary'].",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return self
            if key.split("/")[0] != "streams":
                try:
                    result = _base_getitem("streams/" + key)
                    warnings.warn(
                        f"Key '{key}' not found directly in the BlueskyRun container. "
                        "Trying to access it via the 'streams' namespace for backward-compatibility. "
                        "This behavior is deprecated and will be removed in a future release. "
                        "Please consider migrating the catalog structure to the new layout.",
                        DeprecationWarning,
                        stacklevel=2,
                    )
                    return result
                except KeyError:
                    raise KeyError from e
            elif key.split("/")[0] == "streams":
                try:
                    result = _base_getitem(key[len("streams/") :])
                    warnings.warn(
                        f"Looks like you are trying to access '{key}' via a 'streams' namespace, "
                        "but there is no 'streams' namespace in this BlueskyRun, which follows the new layout. "
                        f"Please access the stream directly, e.g. run['{key}'] instead of run['streams/{key}'].",
                        DeprecationWarning,
                        stacklevel=2,
                    )
                    return result
                except KeyError:
                    raise KeyError from e
            else:
                raise KeyError from e

    def _keys_slice(
        self, start, stop, direction, page_size: int | None = None, **kwargs
    ):
        sorted_keys = (
            reversed(self._stream_names) if direction < 0 else self._stream_names
        )
        return (yield from sorted_keys[start:stop])

    def _items_slice(
        self, start, stop, direction, page_size: int | None = None, **kwargs
    ):
        sorted_keys = (
            reversed(self._stream_names) if direction < 0 else self._stream_names
        )
        for key in sorted_keys[start:stop]:
            yield key, self[key]

    def __iter__(self):
        yield from self._stream_names

    def documents(self, fill=False):
        with io.BytesIO() as buffer:
            self.export(buffer, format="application/json-seq")
            buffer.seek(0)
            for line in buffer:
                parsed = json.loads(line.decode().strip())
                yield parsed["name"], _document_types[parsed["name"]](parsed["doc"])


class BlueskyRunV2SQL(BlueskyRunV2, _BlueskyRunSQL):
    def __getitem__(self, key):
        # For v2, we need to handle the streams and configs keys specially
        if isinstance(key, tuple):
            key = "/".join(key)

        key, *rest = key.split("/", 1)

        if key == "streams":
            raise KeyError(
                "Looks like you are trying to access the 'streams' namespace, "
                "but this pathway has never been supported in the .v2 BlueskyRun client. "
                "Please access the stream directly, e.g. run['primary']."
            )

        stream_composite_client = super().__getitem__(key)
        stream_container = BlueskyEventStreamV2SQL.from_stream_client(
            stream_composite_client
        )

        return stream_container[rest[0]] if rest else stream_container


class BlueskyRunV3(_BlueskyRunSQL):
    """A BlueskyRun that is backed by a SQL database."""

    _version = "3.0"

    def __new__(cls, context, *, item, structure_clients, **kwargs):
        # When inheriting, return the class itself
        if cls is not BlueskyRunV3 or cls._is_sql(item):
            return super().__new__(
                cls, context, item=item, structure_clients=structure_clients, **kwargs
            )
        return BlueskyRunV2Mongo(
            context, item=item, structure_clients=structure_clients, **kwargs
        )

    def __getattr__(self, key):
        # A shortcut to the stream data
        if key in self._stream_names:
            return self["streams"][key] if self._has_streams_namespace else self[key]

        return super().__getattr__(key)

    def __repr__(self):
        metadata = self.metadata
        datetime_ = datetime.fromtimestamp(metadata["start"]["time"])
        return (
            f"<BlueskyRun v{self._version} "
            f"streams: {set(self._stream_names) or 'NONE'} "
            f"scan_id={metadata['start'].get('scan_id', 'UNSET')!s} "  # (scan_id is optional in the schema)
            f"uid={metadata['start']['uid'][:8]!r} "  # truncated uid
            f"{datetime_.isoformat(sep=' ', timespec='minutes')}"
            ">"
        )

    @property
    def v1(self):
        "Access to legacy interface"
        return self.v2.v1

    @property
    def v2(self):
        structure_clients = copy.copy(self.structure_clients)
        structure_clients.set("BlueskyRun", lambda: BlueskyRunV2)
        return BlueskyRunV2(
            self.context, item=self.item, structure_clients=structure_clients
        )

    @property
    def v3(self):
        return self



import json
from datetime import datetime
from pathlib import Path

# NOTE: This code is duplicated in src/bluesky/callbacks/json_writer.py


class JSONWriter:
    """Writer of Bluesky documents of a single run into a JSON file as an array.

    The file is created when a Start document is received, each new document is
    written immediately, and the JSON array is closed when the "stop" document
    is received.
    """

    def __init__(
        self,
        dirname: str,
        filename: str | None = None,
    ):
        self.dirname = Path(dirname)
        self.filename = filename

    def __call__(self, name, doc):
        if name == "start":
            self.filename = self.filename or f"{doc['uid'].split('-')[0]}.json"
            with open(self.dirname / self.filename, "w") as file:
                file.write("[\n")
                json.dump({"name": name, "doc": doc}, file)
                file.write(",\n")

        elif name == "stop":
            with open(self.dirname / self.filename, "a") as file:
                json.dump({"name": name, "doc": doc}, file)
                file.write("\n]")

        else:
            with open(self.dirname / self.filename, "a") as file:
                json.dump({"name": name, "doc": doc}, file)
                file.write(",\n")


class JSONLinesWriter:
    """Writer of Bluesky documents into a JSON Lines file

    If the file already exists, new documents will be appended to it.
    """

    def __init__(self, dirname: str, filename: str | None = None):
        self.dirname = Path(dirname)
        self.filename = filename

    def __call__(self, name, doc):
        if not self.filename:
            if name == "start":
                # If the first document is a start document, use the uid to create a filename
                self.filename = f"{doc['uid'].split('-')[0]}.jsonl"
            else:
                # If the first document is not a start document, use the current date
                self.filename = f"{datetime.today().strftime('%Y-%m-%d')}.jsonl"
        mode = "a" if (self.dirname / self.filename).exists() else "w"

        with open(self.dirname / self.filename, mode) as file:
            json.dump({"name": name, "doc": doc}, file)
            file.write("\n")



"""
This module vendors the Dispatcher from bluesky.run_engine, in order
to avoid a bluesky dependency, since bluesky-tiled-plugins is frequently
used in data analysis environments where a bluesky dependency would be
superfluous.

This dispatcher could in the future be move upstream to event_model
where it could be shared by bluesky and bluesky-tiled-plugins.

That code has been extremely stable for about ten years, so divergence
is not a pressing concern.
"""

import sys
import types
from itertools import count
from warnings import warn
from weakref import WeakKeyDictionary, ref

from event_model import DocumentNames


class Dispatcher:
    """Dispatch documents to user-defined consumers on the main thread."""

    def __init__(self):
        self.cb_registry = CallbackRegistry(allowed_sigs=DocumentNames)
        self._counter = count()
        self._token_mapping = dict()  # noqa: C408

    def process(self, name, doc):
        """
        Dispatch document ``doc`` of type ``name`` to the callback registry.

        Parameters
        ----------
        name : {'start', 'descriptor', 'event', 'stop'}
        doc : dict
        """
        exceptions = self.cb_registry.process(name, name.name, doc)
        for exc, traceback in exceptions:  # noqa: B007
            warn(  # noqa: B028
                "A %r was raised during the processing of a %s "  # noqa: UP031
                "Document. The error will be ignored to avoid "
                "interrupting data collection. To investigate, "
                "set RunEngine.ignore_callback_exceptions = False "
                "and run again." % (exc, name.name)
            )

    def subscribe(self, func, name="all"):
        """
        Register a callback function to consume documents.

        .. versionchanged :: 0.10.0
            The order of the arguments was swapped and the ``name``
            argument has been given a default value, ``'all'``. Because the
            meaning of the arguments is unambiguous (they must be a callable
            and a string, respectively) the old order will be supported
            indefinitely, with a warning.

        .. versionchanged :: 0.10.0
            The order of the arguments was swapped and the ``name``
            argument has been given a default value, ``'all'``. Because the
            meaning of the arguments is unambiguous (they must be a callable
            and a string, respectively) the old order will be supported
            indefinitely, with a warning.

        Parameters
        ----------
        func: callable
            expecting signature like ``f(name, document)``
            where name is a string and document is a dict
        name : {'all', 'start', 'descriptor', 'event', 'stop'}, optional
            the type of document this function should receive ('all' by
            default).

        Returns
        -------
        token : int
            an integer ID that can be used to unsubscribe

        See Also
        --------
        :meth:`Dispatcher.unsubscribe`
            an integer token that can be used to unsubscribe
        """
        if callable(name) and isinstance(func, str):
            name, func = func, name
            warn(  # noqa: B028
                "The order of the arguments has been changed. Because the "
                "meaning of the arguments is unambiguous, the old usage will "
                "continue to work indefinitely, but the new usage is "
                "encouraged: call subscribe(func, name) instead of "
                "subscribe(name, func). Additionally, the 'name' argument "
                "has become optional. Its default value is 'all'."
            )
        if name == "all":
            private_tokens = []
            for key in DocumentNames:
                private_tokens.append(self.cb_registry.connect(key, func))
            public_token = next(self._counter)
            self._token_mapping[public_token] = private_tokens
            return public_token

        name = DocumentNames[name]
        private_token = self.cb_registry.connect(name, func)
        public_token = next(self._counter)
        self._token_mapping[public_token] = [private_token]
        return public_token

    def unsubscribe(self, token):
        """
        Unregister a callback function using its integer ID.

        Parameters
        ----------
        token : int
            the integer ID issued by :meth:`Dispatcher.subscribe`

        See Also
        --------
        :meth:`Dispatcher.subscribe`
        """
        for private_token in self._token_mapping.pop(token, []):
            self.cb_registry.disconnect(private_token)

    def unsubscribe_all(self):
        """Unregister all callbacks from the dispatcher."""
        for public_token in list(self._token_mapping.keys()):
            self.unsubscribe(public_token)

    @property
    def ignore_exceptions(self):
        return self.cb_registry.ignore_exceptions

    @ignore_exceptions.setter
    def ignore_exceptions(self, val):
        self.cb_registry.ignore_exceptions = val


class CallbackRegistry:
    """
    See matplotlib.cbook.CallbackRegistry. This is a simplified since
    ``bluesky`` is python3.4+ only!
    """

    def __init__(self, ignore_exceptions=False, allowed_sigs=None):
        self.ignore_exceptions = ignore_exceptions
        self.allowed_sigs = allowed_sigs
        self.callbacks = dict()  # noqa: C408
        self._cid = 0
        self._func_cid_map = {}

    def __getstate__(self):
        # We cannot currently pickle the callables in the registry, so
        # return an empty dictionary.
        return {}

    def __setstate__(self, state):
        # re-initialise an empty callback registry
        self.__init__()

    def connect(self, sig, func):
        """Register ``func`` to be called when ``sig`` is generated

        Parameters
        ----------
        sig
        func

        Returns
        -------
        cid : int
            The callback index. To be used with ``disconnect`` to deregister
            ``func`` so that it will no longer be called when ``sig`` is
            generated
        """
        if self.allowed_sigs is not None:
            if sig not in self.allowed_sigs:
                raise ValueError(f"Allowed signals are {self.allowed_sigs}")
        self._func_cid_map.setdefault(sig, WeakKeyDictionary())
        # Note proxy not needed in python 3.
        # TODO rewrite this when support for python2.x gets dropped.
        # Following discussion with TC: weakref.WeakMethod can not be used to
        #   replace the custom 'BoundMethodProxy', because it does not accept
        #   the 'destroy callback' as a parameter. The 'destroy callback' is
        #   necessary to automatically unsubscribe CB registry from the callback
        #   when the class object is destroyed and this is the main purpose of
        #   BoundMethodProxy.
        proxy = _BoundMethodProxy(func)
        if proxy in self._func_cid_map[sig]:
            return self._func_cid_map[sig][proxy]

        proxy.add_destroy_callback(self._remove_proxy)
        self._cid += 1
        cid = self._cid
        self._func_cid_map[sig][proxy] = cid
        self.callbacks.setdefault(sig, dict())  # noqa: C408
        self.callbacks[sig][cid] = proxy
        return cid

    def _remove_proxy(self, proxy):
        # need the list because `del self._func_cid_map[sig]` mutates the dict
        for sig, proxies in list(self._func_cid_map.items()):
            try:
                # Here we need to delete the last reference to proxy (in 'self.callbacks[sig]')
                #   The respective entries in 'self._func_cid_map' are deleted automatically,
                #   since 'self._func_cid_map[sig]' entries are WeakKeyDictionary objects.
                del self.callbacks[sig][proxies[proxy]]
            except KeyError:
                pass

            # Remove dictionary items for signals with no assigned callbacks
            if len(self.callbacks[sig]) == 0:
                del self.callbacks[sig]
                del self._func_cid_map[sig]

    def disconnect(self, cid):
        """Disconnect the callback registered with callback id *cid*

        Parameters
        ----------
        cid : int
            The callback index and return value from ``connect``
        """
        for eventname, callbackd in self.callbacks.items():  # noqa: B007
            try:
                # This may or may not remove entries in 'self._func_cid_map'.
                del callbackd[cid]
            except KeyError:
                continue
            else:
                # Look for cid in 'self._func_cid_map' as well. It may still be there.
                for sig, functions in self._func_cid_map.items():  # noqa: B007
                    for function, value in list(functions.items()):
                        if value == cid:
                            del functions[function]
                return

    def process(self, sig, *args, **kwargs):
        """Process ``sig``

        All of the functions registered to receive callbacks on ``sig``
        will be called with ``args`` and ``kwargs``

        Parameters
        ----------
        sig
        args
        kwargs
        """
        if self.allowed_sigs is not None:
            if sig not in self.allowed_sigs:
                raise ValueError(f"Allowed signals are {self.allowed_sigs}")
        exceptions = []
        if sig in self.callbacks:
            for cid, func in list(self.callbacks[sig].items()):  # noqa: B007
                try:
                    func(*args, **kwargs)
                except ReferenceError:
                    self._remove_proxy(func)
                except Exception as e:
                    if self.ignore_exceptions:
                        exceptions.append((e, sys.exc_info()[2]))
                    else:
                        raise
        return exceptions


class _BoundMethodProxy:
    """
    Our own proxy object which enables weak references to bound and unbound
    methods and arbitrary callables. Pulls information about the function,
    class, and instance out of a bound method. Stores a weak reference to the
    instance to support garbage collection.
    @organization: IBM Corporation
    @copyright: Copyright (c) 2005, 2006 IBM Corporation
    @license: The BSD License
    Minor bugfixes by Michael Droettboom
    """

    def __init__(self, cb):
        self._hash = hash(cb)
        self._destroy_callbacks = []
        try:
            # This branch is successful if 'cb' bound method and class method,
            #   but destroy_callback mechanism works only for bound methods,
            #   since cb.__self__ points to class instance only for
            #   bound methods, not for class methods. Therefore destroy_callback
            #   will not be called for class methods.
            try:
                self.inst = ref(cb.__self__, self._destroy)
            except TypeError:
                self.inst = None
            self.func = cb.__func__
            self.klass = cb.__self__.__class__

        except AttributeError:
            # 'cb' is a function, callable object or static method.
            # No weak reference is created, strong reference is stored instead.
            self.inst = None
            self.func = cb
            self.klass = None

    def add_destroy_callback(self, callback):
        self._destroy_callbacks.append(_BoundMethodProxy(callback))

    def _destroy(self, wk):
        for callback in self._destroy_callbacks:
            try:
                callback(self)
            except ReferenceError:
                pass

    def __getstate__(self):
        d = self.__dict__.copy()
        # de-weak reference inst
        inst = d["inst"]
        if inst is not None:
            d["inst"] = inst()
        return d

    def __setstate__(self, statedict):
        self.__dict__ = statedict
        inst = statedict["inst"]
        # turn inst back into a weakref
        if inst is not None:
            self.inst = ref(inst)

    def __call__(self, *args, **kwargs):
        """
        Proxy for a call to the weak referenced object. Take
        arbitrary params to pass to the callable.
        Raises `ReferenceError`: When the weak reference refers to
        a dead object
        """
        if self.inst is not None and self.inst() is None:
            raise ReferenceError
        if self.inst is not None:
            # build a new instance method with a strong reference to the
            # instance

            mtd = types.MethodType(self.func, self.inst())

        else:
            # not a bound method, just return the func
            mtd = self.func
        # invoke the callable and return the result
        return mtd(*args, **kwargs)

    def __eq__(self, other):
        """
        Compare the held function and instance with that held by
        another proxy.
        """
        try:
            if self.inst is None:
                return self.func == other.func and other.inst is None
            return self.func == other.func and self.inst() == other.inst()
        except Exception:
            return False

    def __ne__(self, other):
        """
        Inverse of __eq__.
        """
        return not self.__eq__(other)

    def __hash__(self):
        return self._hash
